"""
    -------------------------------------------------------------------
    sleap_utils.py: General helper functions for interfacing with SLEAP
    -------------------------------------------------------------------
"""

import itertools
import logging
import pathlib
from typing import List, Tuple

import attr
import h5py
import numpy as np
import sleap

logger = logging.getLogger(__name__)


# Pulled from SLEAP's codebase
def get_location_matrix(
    labels: sleap.Labels, all_frames: bool, video: sleap.Video = None
) -> Tuple[np.ndarray, List]:
    """Builds numpy matrix with point location data. Note: This function assumes either all instances have tracks or no instances have tracks.

    Args:
        labels (sleap.Labels): The :py:class:`sleap.Labels` from which to get data.
        all_frames (bool): If True, then includes zeros so that frame index will line up with columns in the output. Otherwise, there will only be columns for the frames between the first and last frames with labeling data.
        video (sleap.Video, optional): The :py:class:`Video` from which to get data. If no `video` is specified, then the first video in `source_object` videos list will be used. If there are no labeled frames in the `video`, then None will be returned. Defaults to None.

    Returns:
        Tuple[np.ndarray, List[str]]: point location array with shape (frames, nodes, 2, tracks) List[str]: list of node names
    """
    # Assumes either all instances have tracks or no instances have tracks
    track_count = len(labels.tracks) or 1
    node_count = len(labels.skeletons[0].nodes)

    # Retrieve frames from current video only
    try:
        if video is None:
            video = labels.videos[0]
    except IndexError as e:
        print("There are no videos in this project. No occupancy matrix to return.")
        raise e
    labeled_frames = labels.get(video)

    frame_idxs = [lf.frame_idx for lf in labeled_frames]
    frame_idxs.sort()

    try:
        first_frame_idx = 0 if all_frames else frame_idxs[0]

        frame_count = (
            frame_idxs[-1] - first_frame_idx + 1
        )  # count should include unlabeled frames
    except IndexError as e:
        print(f"No labeled frames in {video.filename}. No occupancy matrix to return.")
        raise e

    # Desired MATLAB format:
    # "tracks"              frames * nodes * 2 * tracks
    # "track_names"         tracks

    locations_matrix = np.full(
        (frame_count, node_count, 2, track_count), np.nan, dtype=float
    )

    # Assumes either all instances have tracks or no instances have tracks
    # Prefer user-labeled instances over predicted instances
    tracks = labels.tracks or [None]  # Comparator in case of project with no tracks
    lfs_instances = list()
    warning_flag = False
    for lf in labeled_frames:

        user_instances = lf.user_instances
        predicted_instances = lf.predicted_instances
        for track in tracks:
            track_instances = list()
            # If a user-instance exists for this track, then use user-instance
            user_track_instances = [
                inst for inst in user_instances if inst.track == track
            ]
            if len(user_track_instances) > 0:
                track_instances = user_track_instances
            else:
                # Otherwise, if a predicted instance exists, then use the predicted
                predicted_track_instances = [
                    inst for inst in predicted_instances if inst.track == track
                ]
                if len(predicted_track_instances) > 0:
                    track_instances = predicted_track_instances

            lfs_instances.extend([(lf, inst) for inst in track_instances])

            # Set warning flag if more than one instances on a track in a single frame
            warning_flag = warning_flag or (
                (track is not None) and (len(track_instances) > 1)
            )

    if warning_flag:
        print(
            "\nWarning! "
            "There are more than one instances per track on a single frame.\n"
        )

    for lf, inst in lfs_instances:
        frame_i = lf.frame_idx - first_frame_idx
        # Assumes either all instances have tracks or no instances have tracks
        if inst.track is None:
            # We could use lf.instances.index(inst) but then we'd need
            # to calculate the number of "tracks" based on the max number of
            # instances in any frame, so for now we'll assume that there's
            # a single instance if we aren't using tracks.
            track_i = 0
        else:
            track_i = labels.tracks.index(inst.track)

        locations_matrix[frame_i, ..., track_i] = inst.numpy()
    return locations_matrix, labels.skeletons[0].nodes


def load_tracks_from_slp(slp_path: str) -> Tuple[np.ndarray, List[str]]:
    """Loads tracks from a sleap project file or analysis h5

    Args:
        slp_path (str): Path to SLEAP project file or analysis h5. This file is typically generated using `sleap-convert`. Analysis h5s are much faster to load and thus preferable.

    Returns:
        Tuple[np.ndarray, List[str]]: Tuple of (locations_matrix, node_names) where location matrix is a numpy array with shape (frames, nodes, 2, tracks) and node_names is a list of node names.
    """
    if pathlib.Path(slp_path).suffix == ".slp":
        dset = sleap.load_file(slp_path)
        locations, node_names = get_location_matrix(dset, all_frames=True)
        node_names = [node.name for node in node_names]
    elif pathlib.Path(slp_path).suffix == ".h5":
        with h5py.File(slp_path, "r") as f:
            locations = f["tracks"][:].T
            node_names = [n.decode() for n in f["node_names"][:]]
    return locations, node_names


def update_labeled_frames(
    slp_path: str, matching_dict: dict, first_frame_idx: int, last_frame_idx: int
) -> List[sleap.LabeledFrame]:
    """Generates a list of labeled frames from a sleap project file with tracks updated from the provided matching_dict.

    Args:
        slp_path (str): Path to SLEAP project file (.slp) file.
        matching_dict (dict): Dictionary of format [Frame][Track] containing the tag number for each track in each frame.
        first_frame_idx (int): First frame index to include in the reconstructed h5.
        last_frame_idx (int): Last frame index to include in the reconstructed h5.

    Returns:
        List[sleap.LabeledFrame]: List of labeled frames with updated tracks.
    """
    labels = sleap.load_file(slp_path)
    # frames = sorted(labels.labeled_frames, key=operator.attrgetter("frame_idx"))

    # Matching dict is of the form [Frame][Track][Tag]
    # List of lists of tracks
    tags = [track_dict.values() for track_dict in matching_dict.values()]

    # Set of all tags
    tags = set(itertools.chain(*tags))
    logger.info("Total tags: %s", len(tags))
    new_tracks = {
        tag: sleap.Track(spawned_on=0, name=f"ArUcoTag#{tag}") for tag in tags
    }
    new_lfs = []

    # Run tracking on every frame
    for lf in labels.labeled_frames[first_frame_idx : last_frame_idx + 1]:
        tracked_instances = []
        for inst in lf.instances:
            try:
                inst_name = int(inst.track.name.split("_")[-1])
                track = matching_dict[lf.frame_idx][inst_name]
                tracked_instances.append(attr.evolve(inst, track=new_tracks[track]))
            except Exception as e:
                tracked_instances.append(attr.evolve(inst, track=None))
                continue
        new_lfs.append(attr.evolve(lf, instances=tracked_instances))
    tracked_labels = sleap.Labels(new_lfs)
    return tracked_labels



# Basic Information
__name__ = "naps-track"
__version__ = "1.0.2"
__summary__ = "NAPS (NAPS is ArUco Plus SLEAP)"
__url__ = "https://github.com/kocherlab/naps"
__code__ = "https://github.com/kocherlab/naps"
__issue__ = "https://github.com/kocherlab/naps/issues"
__docs__ = "https://naps.readthedocs.io/en/latest/"
__license__ = "MIT"
__copyright__ = "2022"

# Author Information
__authors__ = "Scott Wolf, Dee Ruttenberg, Daniel Knapp, Andrew Webb, Ian Traniello, Grace McKenzie-Smith, Joshua Shaevitz, and Sarah Kocher"
__email__ = "22243650+Wolfffff@users.noreply.github.com, 19213578+aewebb80@users.noreply.github.com"

import logging

# Start
import sys

logging.basicConfig(
    stream=sys.stdout,
    format="%(asctime)s %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)



#!/usr/bin/env python
import os
from collections import defaultdict
from typing import Callable

import cv2
import numpy as np

from naps.cost_matrix import CostMatrix


class Matching:
    """Matching pipeline.

    Attributes:
        video_filename (str): Matching video filename (i.e. path).
        video_first_frame (str): Frame to start matching.
        video_last_frame (str): Frame to end matching.
        marker_detector (Callable): Function used to detect/assign markers.
        aruco_crop_size (int): Crop size used for marker detection.
        half_rolling_window_size (int): Window size (upstream/downstream) used by the cost matrix.
        tag_node_dict (dict): Dictionary of frames, tracks, and node coordinates.
        threads (int): The number of CPU threads to use.
        min_sleap_score (float): Minimum sleap score required for matching. Should this be removed?
    """

    def __init__(
        self,
        video_filename: str,
        video_first_frame: int,
        video_last_frame: int,
        marker_detector: Callable,
        aruco_crop_size: int,
        half_rolling_window_size: int,
        tag_node_dict: dict,
        min_sleap_score: float = 0.1,
        **kwargs,
    ):

        # Confirm the video file exists
        if not os.path.isfile(video_filename):
            raise Exception(f"{video_filename} does not exist")

        # Video arguments
        self.video_filename = video_filename
        self.video_first_frame = video_first_frame
        self.video_last_frame = video_last_frame

        # SLEAP arguments
        self.tag_node_dict = tag_node_dict
        self.min_sleap_score = min_sleap_score

        # Matching arguments
        self.half_rolling_window_size = half_rolling_window_size
        self.aruco_crop_size = aruco_crop_size
        self.marker_detector = marker_detector

    def match(self) -> defaultdict(lambda: defaultdict(str)):
        """Performs matching.

        Returns:
            defaultdict(lambda: defaultdict(str)): Dictionary of matching results with the form dictionary[frame][track] = tag.
        """

        # Create a dict to store the matches for this job
        job_match_dict = {}

        # Set the current frame of the job
        current_frame = self.video_first_frame
        print(f"Processing frames {self.video_first_frame} to {self.video_last_frame}")

        # Initialize OpenCV, then set the starting frame (0-based)
        video = cv2.VideoCapture(self.video_filename)
        video.set(cv2.CAP_PROP_POS_FRAMES, current_frame)

        # Read in the frame, confirm it was successful
        frame = MatchFrame.fromCV2(*video.read())
        while frame and current_frame <= self.video_last_frame:

            # Crop and return the ArUco tags
            frame.cropMarkerWithCoordsArray(
                self.tag_node_dict[current_frame], self.aruco_crop_size
            )
            job_match_dict[current_frame] = frame.returnMarkerTags(self.marker_detector)

            # Advance to the next frame
            frame = MatchFrame.fromCV2(*video.read())
            current_frame += 1

        # Create the cost matrix and assign the track/tag pairs for each frame
        job_cost_matrix = CostMatrix.fromDict(
            job_match_dict,
            self.video_first_frame,
            self.video_last_frame,
            self.half_rolling_window_size,
        )
        return job_cost_matrix.assignTrackTagPairs()


class MatchFrame:
    """Class used to assign markers for a frame and associated data.

    Attributes:
        frame (np.ndarray): Value array of the frame image.
        frame_exists (bool): Boolean indicating if the frame was read correctly.
        frame_images (dict): Dictionary to store value arrays for each marker image.

    """

    def __init__(self, frame_exists: bool, frame: np.ndarray, *args, **kwargs):

        # Image arguments
        self.frame = frame
        self.frame_exists = frame_exists
        self.frame_images = {}

    def __nonzero__(self):
        """Replaces nonzero for the class

        Returns:
            bool: Boolean indicating if the frame was read correctly.
        """

        return self.frame_exists

    @classmethod
    def fromCV2(cls, *args, **kwargs):
        """Class methods to create a MatchFrame from CV2.read()

        Args:
            bool: Boolean indicating if the frame was read correctly.
            np.ndarray: Value array of the frame image.

        Returns:
            MatchFrame: A MatchFrame object.
        """

        return cls(*args, **kwargs)

    def cropMarkerWithCoordsArray(self, coords_dict, crop_size: int):
        """Creates cropped images for each marker in the coords_dict.

        Args:
            coords_dict (dict): Dictionary of marker coordinates to crop for each track.
            crop_size (float): Crop size.

        """

        def croppedCoords(coord: float, crop_size: float, coord_max: int):
            """Gets the cropped coordinates for the given single coordinate

            Args:
                coord (float): Coordinate to get crop range.
                crop_size (float): Crop size.
                coord_max (int): Maximum coordinate possible before leaving frame.

            Returns:
                tuple(int): The minimum and maximum coordinates for the given single coordinate

            """
            return np.maximum(int(coord) - crop_size, 0), np.minimum(
                int(coord) + crop_size, coord_max - 1
            )

        # Loop the frame track coordinates
        for track, coords in coords_dict.items():

            # Skip track if NaN found in coordinates
            if np.isnan(coords).any():
                self.frame_images[track] = None
                continue

            # Assign the min/max coords for cropping
            y_min, y_max = croppedCoords(coords[1], crop_size, self.frame.shape[0])
            x_min, x_max = croppedCoords(coords[0], crop_size, self.frame.shape[1])

            # Assign and store the cropped track image
            self.frame_images[track] = self.frame[y_min:y_max, x_min:x_max, 0]

    def returnMarkerTags(self, marker_detect: Callable):
        """Detect marker tags using the specified marker_detect function

        Args:
            marker_detect (Callable): Marker detector function. Function is expected
            to return a list of marker assignments.

        Returns:
            defaultdict(list): Dictionary of matching results for a single frame.
        """

        track_tag_dict = defaultdict(list)

        # Loop the frame track images
        for track, track_image in self.frame_images.items():

            # Store None if no image was found
            if track_image is None:
                track_tag_dict[track].append(None)
                continue

            # Assign the marker using the detector function
            track_tag_dict[track].extend(marker_detect(track_image))

        return track_tag_dict



#!/usr/bin/env python
from collections import defaultdict

import pandas as pd
from scipy.optimize import linear_sum_assignment


class CostMatrix:
    def __init__(
        self,
        unmatched_dict: dict,
        first_frame: int,
        last_frame: int,
        half_rolling_window_size: int,
        **kwargs
    ):

        # Matrix arguments
        self.unmatched_dict = unmatched_dict

        # Matching arguments
        self.first_frame = first_frame
        self.last_frame = last_frame
        self.half_rolling_window_size = half_rolling_window_size

        # Assignment argument
        self.assignment_method = self._linearAssignment

    @classmethod
    def fromDict(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def assignTrackTagPairs(self):

        # Create the cost matrix
        cost_dict = defaultdict(lambda: defaultdict(int))
        matched_dict = defaultdict(lambda: defaultdict(str))

        # Loop the frames in the unmatched matrix
        for frame in range(self.first_frame, self.last_frame + 1):

            # Assign the track/tag combinations for this frame
            track_tag_dict = self.unmatched_dict[frame]
            for track, tag_list in track_tag_dict.items():
                for tag in tag_list:
                    if tag is not None:
                        cost_dict[track][tag] -= 1

            # Assign the rolling window frame
            match_frame = frame - self.half_rolling_window_size

            # Check if the match frame should undergo linear assignment
            if (
                match_frame < self.first_frame + self.half_rolling_window_size
                or match_frame > self.last_frame - self.half_rolling_window_size
            ):
                continue

            # Assign the first frame of the window
            start_of_window = match_frame - self.half_rolling_window_size

            # Remove a frame, if needed
            if start_of_window > self.first_frame:
                track_tag_dict = self.unmatched_dict[start_of_window - 1]
                for track, tag_list in track_tag_dict.items():
                    for tag in tag_list:
                        if tag != -1:
                            cost_dict[track][tag] += 1
                            if cost_dict[track][tag] == 0:
                                del cost_dict[track][tag]

            # Create a dataframe of the matrix
            cost_dataframe = pd.DataFrame.from_dict(cost_dict).fillna(0)
            cost_dataframe = cost_dataframe.loc[~(cost_dataframe == 0).all(axis=1)]
            cost_dataframe = cost_dataframe.loc[:, ~(cost_dataframe == 0).all(axis=0)]

            # Store the assignments
            for track_index, tag_index in self.assignment_method(cost_dataframe.values):
                matched_dict[match_frame][
                    cost_dataframe.columns[track_index]
                ] = cost_dataframe.index[tag_index]

        for frame in range(self.first_frame + 1, self.last_frame + 1):
            for track, tag in matched_dict[frame - 1].items():
                if (
                    track in self.unmatched_dict[frame].keys()
                    and tag not in matched_dict[frame].values()
                ):
                    matched_dict[frame][track] = matched_dict[frame - 1][track]
        return matched_dict

    @staticmethod
    def _linearAssignment(value_array):
        # Solve the linear assignment problem using Jonker-Volgenant algorithm
        tag_indices, track_indices = linear_sum_assignment(value_array)

        # Store the assignments
        for track_index, tag_index in zip(track_indices, tag_indices):
            yield track_index, tag_index



#!/usr/bin/env python
import argparse
import logging
import time
from collections import defaultdict

import sleap

from naps.aruco import ArUcoModel
from naps.matching import Matching
from naps.sleap_utils import update_labeled_frames

logger = logging.getLogger("NAPS Logger")


def build_parser():
    """Builds the argument parser for the main function.

    Returns:
        argparse.ArgumentParser: Parser for command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="NAPS -- Hybrid tracking using SLEAP and ArUco tags"
    )

    parser.add_argument(
        "--slp-path",
        help="The filepath of the SLEAP (.slp or .h5) file to pull coordinates from. This should correspond with the input video file",
        type=str,
        required=True,
    )
    parser.add_argument(
        "--h5-path",
        help="The filepath of the analysis h5 file to pull coordinates from. This should correspond with the input video file and slp file",
        type=str,
        default=None,
    )

    parser.add_argument(
        "--video-path",
        help="The filepath of the video used with SLEAP",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--tag-node-name",
        help="The ArUco tag SLEAP node name",
        type=str,
        required=True,
        default="tag",
    )

    parser.add_argument(
        "--start-frame",
        help="The zero-based fully-closed frame to begin NAPS assignment",
        type=int,
        required=True,
    )

    parser.add_argument(
        "--end-frame",
        help="The zero-based fully-closed frame to stop NAPS assignment",
        type=int,
        required=True,
    )

    parser.add_argument(
        "--half-rolling-window-size",
        help="Specifies the number of flanking frames (prior and subsequent) required in the rolling window for Hungarian matching a frame",
        type=int,
        default=5,
    )

    parser.add_argument(
        "--aruco-marker-set",
        help="The ArUco markers used in the video",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--aruco-crop-size",
        help="The number of pixels horizontally and vertically around the aruco SLEAP node to identify the marker",
        type=int,
        default=50,
    )

    parser.add_argument(
        "--aruco-adaptive-thresh-win-size-min",
        dest="adaptiveThreshWinSizeMin",
        help="Specifies the value for adaptiveThreshWinSizeMin",
        type=int,
        default=10,
    )

    parser.add_argument(
        "--aruco-adaptive-thresh-win-size-max",
        dest="adaptiveThreshWinSizeMax",
        help="Specifies the value for adaptiveThreshWinSizeMax",
        type=int,
        default=30,
    )

    parser.add_argument(
        "--aruco-adaptive-thresh-win-size-step",
        dest="adaptiveThreshWinSizeStep",
        help="Specifies the value for adaptiveThreshWinSizeStep",
        type=int,
        default=12,
    )

    parser.add_argument(
        "--aruco-adaptive-thresh-constant",
        dest="adaptiveThreshConstant",
        help="Specifies the value for adaptiveThreshConstant",
        type=float,
        default=3,
    )

    parser.add_argument(
        "--aruco-perspective-rm-ignored-margin",
        dest="perspectiveRemoveIgnoredMarginPerCell",
        help="Specifies the value for perspectiveRemoveIgnoredMarginPerCell",
        type=float,
        default=0.13,
    )

    parser.add_argument(
        "--aruco-error-correction-rate",
        dest="errorCorrectionRate",
        help="Specifies the value for errorCorrectionRate",
        type=float,
        default=1,
    )

    parser.add_argument(
        "--output-path",
        help="Output path of the resulting SLEAP analysis file.",
        type=str,
        default="output.analysis.h5",
    )

    return parser


def main(argv=None):
    """Main function for the NAPS tracking script."""

    # Set the start time for the entire pipeline
    t0_total = time.time()

    # Build the arguments from the .parse_args(args)
    args = build_parser().parse_args(argv)

    # Assign the h5 path if not specified
    if args.h5_path is None:
        args.h5_path = args.slp_path

    # Create a track array from the SLEAP file(s)
    logger.info("Loading predictions...")
    t0 = time.time()
    # locations, node_names = load_tracks_from_slp(args.h5_path)
    tag_locations_dict = defaultdict(lambda: defaultdict(tuple))
    labels = sleap.Labels.load_file(args.slp_path)
    for lf in labels.labeled_frames:
        if lf.frame_idx < args.start_frame or lf.frame_idx > args.end_frame:
            continue
        for instance in lf.instances:
            tag_idx = instance.skeleton.node_names.index(args.tag_node_name)
            track_name = int(instance.track.name.split("_")[-1])
            # print(f'Frame {lf.frame_idx} track {track_name} has points {instance.numpy()[tag_idx]}')
            tag_locations_dict[lf.frame_idx][track_name] = instance.numpy()[tag_idx]

    # Create an ArUcoModel with the default/specified parameters
    logger.info("Create ArUco model...")
    t0 = time.time()
    aruco_model = ArUcoModel.withTagSet(
        args.aruco_marker_set,
        adaptiveThreshWinSizeMin=args.adaptiveThreshWinSizeMin,
        adaptiveThreshWinSizeMax=args.adaptiveThreshWinSizeMax,
        adaptiveThreshWinSizeStep=args.adaptiveThreshWinSizeStep,
        adaptiveThreshConstant=args.adaptiveThreshConstant,
        perspectiveRemoveIgnoredMarginPerCell=args.perspectiveRemoveIgnoredMarginPerCell,
        errorCorrectionRate=args.errorCorrectionRate,
    )
    logger.info("ArUco model built in %s seconds.", time.time() - t0)

    # Match the track to the ArUco markers
    logger.info("Starting matching...")
    t0 = time.time()
    matching = Matching(
        args.video_path,
        args.start_frame,
        args.end_frame,
        marker_detector=aruco_model.detect,
        aruco_crop_size=args.aruco_crop_size,
        half_rolling_window_size=args.half_rolling_window_size,
        tag_node_dict=tag_locations_dict,
    )
    matching_dict = matching.match()
    logger.info("Done matching in %s seconds.", time.time() - t0)

    # Create the output
    logger.info("Reconstructing SLEAP file...")
    t0 = time.time()
    # Right now the reconstruction assumes that we each track has a single track ID assigned to it. We'll generalize so that a track can switch IDs over time.
    labels = update_labeled_frames(
        args.slp_path, matching_dict, args.start_frame, args.end_frame
    )
    labels.save(args.output_path)
    logger.info("Done reconstructing SLEAP file in %s seconds.", time.time() - t0)

    logger.info("Complete NAPS runtime: %s", time.time() - t0_total)


if __name__ == "__main__":
    main()



import io
import os

from setuptools import setup

import naps

with io.open("README.rst", "rt", encoding="utf8") as f:
    readme = f.read()

# Read requirements.txt
lib_folder = os.path.dirname(os.path.realpath(__file__))
requirement_path = lib_folder + "/requirements.txt"
requirements = []
if os.path.isfile(requirement_path):
    with open(requirement_path) as f:
        requirements = f.read().splitlines()

# Executable scripts in the package
tool_scripts = ["naps/naps_track.py"]

setup(
    name=naps.__name__,
    version=naps.__version__,
    project_urls={
        "Documentation": naps.__docs__,
        "Code": naps.__code__,
        "Issue tracker": naps.__issue__,
    },
    license=naps.__license__,
    url=naps.__url__,
    description=naps.__summary__,
    long_description_content_type="text/x-rst",
    long_description=readme,
    packages=["naps"],
    install_requires=requirements,
    scripts=tool_scripts,
    entry_points={
        "console_scripts": ["naps-track=naps.naps_track:main"],
    },
    python_requires=">=3.6",
)



#!/usr/bin/env python
import cv2.aruco


class ArUcoModel:
    """Class providing a wrapper around the cv2.aruco library"""

    def __init__(
        self,
        tag_set: str,
        adaptiveThreshWinSizeMin: int,
        adaptiveThreshWinSizeMax: int,
        adaptiveThreshWinSizeStep: int,
        adaptiveThreshConstant: float,
        perspectiveRemoveIgnoredMarginPerCell: float,
        errorCorrectionRate: float,
        **kwargs,
    ):

        # Store the tag set
        self.tag_set = tag_set

        # Store the ArUco parameters
        self.adaptiveThreshWinSizeMin = adaptiveThreshWinSizeMin
        self.adaptiveThreshWinSizeMax = adaptiveThreshWinSizeMax
        self.adaptiveThreshWinSizeStep = adaptiveThreshWinSizeStep
        self.adaptiveThreshConstant = adaptiveThreshConstant
        self.errorCorrectionRate = errorCorrectionRate
        self.perspectiveRemoveIgnoredMarginPerCell = (
            perspectiveRemoveIgnoredMarginPerCell
        )

        """
        Set the ArUco dict and params to None. Create w/ buildModel
        as we cannot pickle them when using multiprocessing
        """
        self.aruco_dict = None
        self.aruco_params = None
        self.model_built = False

    @classmethod
    def withTagSet(cls, tag_set, **kwargs):
        return cls(tag_set, **kwargs)

    def build(self):

        # Assign the aruco dict
        self.aruco_dict = self._assignArucoDict(self.tag_set)

        """
        ArUco parameters:
        These have been adjusted by dyknapp but are worth playing with if ArUco is too slow or not detecting enough tags.
        These thresholding parameters DRAMATICALLY improve detection rate, while DRAMATICALLY hurting performance.
        Since super fast processing isn't really necessary here they should be fine as is.
        """
        self.aruco_params = cv2.aruco.DetectorParameters_create()

        """
        Assign the corner refinement method:

        Should we permit all options?
        """
        self.aruco_params.cornerRefinementMethod = cv2.aruco.CORNER_REFINE_SUBPIX

        # Window parameters
        self.aruco_params.adaptiveThreshWinSizeMin = self.adaptiveThreshWinSizeMin
        self.aruco_params.adaptiveThreshWinSizeMax = self.adaptiveThreshWinSizeMax
        self.aruco_params.adaptiveThreshWinSizeStep = self.adaptiveThreshWinSizeStep

        # If too slow, start by adjusting this one up. If we want more tags, lower it (diminishing returns)
        self.aruco_params.adaptiveThreshConstant = self.adaptiveThreshConstant

        # No note for this option
        self.aruco_params.perspectiveRemoveIgnoredMarginPerCell = (
            self.perspectiveRemoveIgnoredMarginPerCell
        )

        # If false positives are a problem, lower this parameter.
        self.aruco_params.errorCorrectionRate = self.errorCorrectionRate

        # Indicate the model has been built
        self.model_built = True

    def detect(self, img):

        # Build the model if needed
        if not self.model_built:
            self.build()

        # Detect ArUco tag(s) within the image
        corners, tags, _ = cv2.aruco.detectMarkers(
            img, self.aruco_dict, parameters=self.aruco_params
        )

        # Return None if no ArUco tag was found
        if len(corners) == 0:
            return [None]

        # Return detected ArUco tags
        return [marker_tag[0] for _, marker_tag in zip(corners, tags)]

    def _assignArucoDict(self, tag_set):

        # Define names of each possible ArUco tag OpenCV supports
        ARUCO_DICT = {
            "DICT_4X4_50": cv2.aruco.DICT_4X4_50,
            "DICT_4X4_100": cv2.aruco.DICT_4X4_100,
            "DICT_4X4_250": cv2.aruco.DICT_4X4_250,
            "DICT_4X4_1000": cv2.aruco.DICT_4X4_1000,
            "DICT_5X5_50": cv2.aruco.DICT_5X5_50,
            "DICT_5X5_100": cv2.aruco.DICT_5X5_100,
            "DICT_5X5_250": cv2.aruco.DICT_5X5_250,
            "DICT_5X5_1000": cv2.aruco.DICT_5X5_1000,
            "DICT_6X6_50": cv2.aruco.DICT_6X6_50,
            "DICT_6X6_100": cv2.aruco.DICT_6X6_100,
            "DICT_6X6_250": cv2.aruco.DICT_6X6_250,
            "DICT_6X6_1000": cv2.aruco.DICT_6X6_1000,
            "DICT_7X7_50": cv2.aruco.DICT_7X7_50,
            "DICT_7X7_100": cv2.aruco.DICT_7X7_100,
            "DICT_7X7_250": cv2.aruco.DICT_7X7_250,
            "DICT_7X7_1000": cv2.aruco.DICT_7X7_1000,
            "DICT_ARUCO_ORIGINAL": cv2.aruco.DICT_ARUCO_ORIGINAL,
            "DICT_APRILTAG_16h5": cv2.aruco.DICT_APRILTAG_16h5,
            "DICT_APRILTAG_25h9": cv2.aruco.DICT_APRILTAG_25h9,
            "DICT_APRILTAG_36h10": cv2.aruco.DICT_APRILTAG_36h10,
            "DICT_APRILTAG_36h11": cv2.aruco.DICT_APRILTAG_36h11,
        }

        # Check there are no problem with the tag set
        if tag_set not in ARUCO_DICT:
            raise Exception(f"Unable to assign tag set: {tag_set}")

        # Return the OpenCV tags
        return cv2.aruco.Dictionary_get(ARUCO_DICT[tag_set])
